@@ -50,6 +50,9 @@ type Harness interface {
 	// same storage bucket; i.e., a blob created using one driver.Bucket must
 	// be readable by a subsequent driver.Bucket.
 	MakeDriver(ctx context.Context) (driver.Bucket, error)
+	// MakeDriverForNonexistentBucket creates a driver.Bucket for a nonexistent
+	// bucket. If that concept doesn't make sense for a driver, return (nil, nil).
+	MakeDriverForNonexistentBucket(ctx context.Context) (driver.Bucket, error)
 	// HTTPClient should return an unauthorized *http.Client, or nil.
 	// Required if the service supports SignedURL.
 	HTTPClient() *http.Client
@@ -75,6 +78,8 @@ type HarnessMaker func(ctx context.Context, t *testing.T) (Harness, error)
 //    on the single blob entry returned.
 // 7. Tries to read a non-existent blob, and calls ErrorCheck with the error.
 // 8. Makes a copy of the blob, using BeforeCopy as a CopyOption.
+// 9. Calls SignedURL using BeforeSign as a SignedURLOption for each supported
+//    signing method (i.e. GET, PUT and DELETE).
 //
 // For example, an AsTest might set a driver-specific field to a custom
 // value in BeforeWrite, and then verify the custom value was returned in
@@ -98,6 +103,9 @@ type AsTest interface {
 	// BeforeList will be passed directly to ListOptions as part of listing the
 	// test blob.
 	BeforeList(as func(interface{}) bool) error
+	// BeforeSign will be passed directly to SignedURLOptions as part of
+	// generating a signed URL to the test blob.
+	BeforeSign(as func(interface{}) bool) error
 	// AttributesCheck will be called after fetching the test blob's attributes.
 	// It should call attrs.As and verify the results.
 	AttributesCheck(attrs *blob.Attributes) error
@@ -160,6 +168,13 @@ func (verifyAsFailsOnNil) BeforeList(as func(interface{}) bool) error {
 	return nil
 }
 
+func (verifyAsFailsOnNil) BeforeSign(as func(interface{}) bool) error {
+	if as(nil) {
+		return errors.New("want BeforeSign's As to return false when passed nil")
+	}
+	return nil
+}
+
 func (verifyAsFailsOnNil) AttributesCheck(attrs *blob.Attributes) error {
 	if attrs.As(nil) {
 		return errors.New("want Attributes.As to return false when passed nil")
@@ -183,6 +198,9 @@ func (verifyAsFailsOnNil) ListObjectCheck(o *blob.ListObject) error {
 
 // RunConformanceTests runs conformance tests for driver implementations of blob.
 func RunConformanceTests(t *testing.T, newHarness HarnessMaker, asTests []AsTest) {
+	t.Run("TestNonexistentBucket", func(t *testing.T) {
+		testNonexistentBucket(t, newHarness)
+	})
 	t.Run("TestList", func(t *testing.T) {
 		testList(t, newHarness)
 	})
@@ -248,6 +266,54 @@ func RunBenchmarks(b *testing.B, bkt *blob.Bucket) {
 	})
 }
 
+// testNonexistentBucket tests the functionality of IsAccessible.
+func testNonexistentBucket(t *testing.T, newHarness HarnessMaker) {
+	ctx := context.Background()
+	h, err := newHarness(ctx, t)
+	if err != nil {
+		t.Fatal(err)
+	}
+	defer h.Close()
+
+	// Get a driver instance pointing to a nonexistent bucket.
+	{
+		drv, err := h.MakeDriverForNonexistentBucket(ctx)
+		if err != nil {
+			t.Fatal(err)
+		}
+		if drv == nil {
+			// No such thing as a "nonexistent bucket" for this driver.
+			t.Skip()
+		}
+		b := blob.NewBucket(drv)
+		defer b.Close()
+		exists, err := b.IsAccessible(ctx)
+		if err != nil {
+			t.Fatal(err)
+		}
+		if exists {
+			t.Error("got IsAccessible true for nonexistent bucket, want false")
+		}
+	}
+
+	// Verify that IsAccessible returns true for a real bucket.
+	{
+		drv, err := h.MakeDriver(ctx)
+		if err != nil {
+			t.Fatal(err)
+		}
+		b := blob.NewBucket(drv)
+		defer b.Close()
+		exists, err := b.IsAccessible(ctx)
+		if err != nil {
+			t.Fatal(err)
+		}
+		if !exists {
+			t.Error("got IsAccessible false for real bucket, want true")
+		}
+	}
+}
+
 // testList tests the functionality of List.
 func testList(t *testing.T, newHarness HarnessMaker) {
 	const keyPrefix = "blob-for-list"
@@ -1133,25 +1199,44 @@ func testAttributes(t *testing.T, newHarness HarnessMaker) {
 	if r.ContentType() != contentType {
 		t.Errorf("got Reader.ContentType() %q want %q", r.ContentType(), contentType)
 	}
+	if !a.CreateTime.IsZero() {
+		if a.CreateTime.After(a.ModTime) {
+			t.Errorf("CreateTime %v is after ModTime %v", a.CreateTime, a.ModTime)
+		}
+	}
+	if a.ModTime.IsZero() {
+		t.Errorf("ModTime not set")
+	}
 	if a.Size != int64(len(content)) {
 		t.Errorf("got Size %d want %d", a.Size, len(content))
 	}
 	if r.Size() != int64(len(content)) {
 		t.Errorf("got Reader.Size() %d want %d", r.Size(), len(content))
 	}
+	if a.ETag == "" {
+		t.Error("ETag not set")
+	}
+	// Some basic syntax checks on ETag based on https://en.wikipedia.org/wiki/HTTP_ETag.
+	// It should be of the form "xxxx" or W/"xxxx".
+	if !strings.HasPrefix(a.ETag, "W/\"") && !strings.HasPrefix(a.ETag, "\"") {
+		t.Errorf("ETag should start with W/\" or \" (got %s)", a.ETag)
+	}
+	if !strings.HasSuffix(a.ETag, "\"") {
+		t.Errorf("ETag should end with \" (got %s)", a.ETag)
+	}
 	r.Close()
 
-	t1 := a.ModTime
+	// Modify and re-fetch attributes.
 	if err := b.WriteAll(ctx, key, content, nil); err != nil {
 		t.Fatal(err)
 	}
+
 	a2, err := b.Attributes(ctx, key)
 	if err != nil {
 		t.Errorf("failed Attributes#2: %v", err)
 	}
-	t2 := a2.ModTime
-	if t2.Before(t1) {
-		t.Errorf("ModTime %v is before %v", t2, t1)
+	if a2.ModTime.Before(a.ModTime) {
+		t.Errorf("ModTime %v is before %v", a2.ModTime, a.ModTime)
 	}
 }
 
@@ -1740,7 +1825,14 @@ func testCopy(t *testing.T, newHarness HarnessMaker) {
 		if err != nil {
 			t.Fatal(err)
 		}
-		wantAttr.ModTime = time.Time{} // don't compare this field
+
+		// Clear uncomparable fields.
+		clearUncomparableFields := func(a *blob.Attributes) {
+			a.CreateTime = time.Time{}
+			a.ModTime = time.Time{}
+			a.ETag = ""
+		}
+		clearUncomparableFields(wantAttr)
 
 		// Create another blob that we're going to overwrite.
 		if err := b.WriteAll(ctx, dstKeyExists, []byte("clobber me"), nil); err != nil {
@@ -1764,7 +1856,7 @@ func testCopy(t *testing.T, newHarness HarnessMaker) {
 		if err != nil {
 			t.Fatal(err)
 		}
-		gotAttr.ModTime = time.Time{} // don't compare this field
+		clearUncomparableFields(gotAttr)
 		if diff := cmp.Diff(gotAttr, wantAttr, cmpopts.IgnoreUnexported(blob.Attributes{})); diff != "" {
 			t.Errorf("got %v want %v diff %s", gotAttr, wantAttr, diff)
 		}
@@ -1787,7 +1879,7 @@ func testCopy(t *testing.T, newHarness HarnessMaker) {
 		if err != nil {
 			t.Fatal(err)
 		}
-		gotAttr.ModTime = time.Time{} // don't compare this field
+		clearUncomparableFields(gotAttr)
 		if diff := cmp.Diff(gotAttr, wantAttr, cmpopts.IgnoreUnexported(blob.Attributes{})); diff != "" {
 			t.Errorf("got %v want %v diff %s", gotAttr, wantAttr, diff)
 		}
@@ -2055,7 +2147,7 @@ func testSignedURL(t *testing.T, newHarness HarnessMaker) {
 		t.Error("got nil error, expected error for negative SignedURLOptions.Expiry")
 	}
 
-	// Generate real signed URLs for GET, GET with the query params remvoed, PUT, and DELETE.
+	// Generate a signed URL for GET.
 	getURL, err := b.SignedURL(ctx, key, nil)
 	if err != nil {
 		if gcerrors.Code(err) == gcerrors.Unimplemented {
@@ -2066,6 +2158,7 @@ func testSignedURL(t *testing.T, newHarness HarnessMaker) {
 	} else if getURL == "" {
 		t.Fatal("got empty GET url")
 	}
+
 	// Copy getURL, but remove all query params. This URL should not be allowed
 	// to GET since the client is unauthorized.
 	getURLNoParamsURL, err := url.Parse(getURL)
@@ -2078,6 +2171,8 @@ func testSignedURL(t *testing.T, newHarness HarnessMaker) {
 		allowedContentType   = "text/plain"
 		differentContentType = "application/octet-stream"
 	)
+
+	// Generate a signed URL for PUT, with a non-empty ContentType.
 	putURLWithContentType, err := b.SignedURL(ctx, key, &blob.SignedURLOptions{
 		Method:      http.MethodPut,
 		ContentType: allowedContentType,
@@ -2089,6 +2184,8 @@ func testSignedURL(t *testing.T, newHarness HarnessMaker) {
 	} else if putURLWithContentType == "" {
 		t.Fatal("got empty PUT url")
 	}
+
+	// Generate a signed URL for PUT, with an empty ContentType that's enforced.
 	putURLEnforcedAbsentContentType, err := b.SignedURL(ctx, key, &blob.SignedURLOptions{
 		Method:                   http.MethodPut,
 		EnforceAbsentContentType: true,
@@ -2100,6 +2197,8 @@ func testSignedURL(t *testing.T, newHarness HarnessMaker) {
 	} else if putURLEnforcedAbsentContentType == "" {
 		t.Fatal("got empty PUT url")
 	}
+
+	// Same as above, but not enforced.
 	putURLWithoutContentType, err := b.SignedURL(ctx, key, &blob.SignedURLOptions{
 		Method: http.MethodPut,
 	})
@@ -2108,6 +2207,8 @@ func testSignedURL(t *testing.T, newHarness HarnessMaker) {
 	} else if putURLWithoutContentType == "" {
 		t.Fatal("got empty PUT url")
 	}
+
+	// Generate a signed URL for DELETE.
 	deleteURL, err := b.SignedURL(ctx, key, &blob.SignedURLOptions{Method: http.MethodDelete})
 	if err != nil {
 		t.Fatal(err)
@@ -2123,42 +2224,49 @@ func testSignedURL(t *testing.T, newHarness HarnessMaker) {
 	// PUT the blob. Try with all URLs, only putURL should work when given the
 	// content type used in the signature.
 	type signedURLTest struct {
-		urlMethod   string
-		contentType string
-		url         string
-		wantSuccess bool
+		urlDescription string
+		url            string
+		contentType    string
+		wantSuccess    bool
 	}
 	tests := []signedURLTest{
-		{http.MethodGet, "", getURL, false},
-		{http.MethodDelete, "", deleteURL, false},
+		{"getURL", getURL, "", false},
+		{"deleteURL", deleteURL, "", false},
 	}
 	if putURLWithContentType != "" {
-		tests = append(tests, signedURLTest{http.MethodPut, allowedContentType, putURLWithContentType, true})
-		tests = append(tests, signedURLTest{http.MethodPut, differentContentType, putURLWithContentType, false})
-		tests = append(tests, signedURLTest{http.MethodPut, "", putURLWithContentType, false})
+		tests = append(tests, signedURLTest{"putURLWithContentType", putURLWithContentType, allowedContentType, true})
+		tests = append(tests, signedURLTest{"putURLWithContentType", putURLWithContentType, differentContentType, false})
+		tests = append(tests, signedURLTest{"putURLWithContentType", putURLWithContentType, "", false})
 	}
+	/*
+	 */
 	if putURLEnforcedAbsentContentType != "" {
-		tests = append(tests, signedURLTest{http.MethodPut, "", putURLWithoutContentType, true})
-		tests = append(tests, signedURLTest{http.MethodPut, differentContentType, putURLWithoutContentType, false})
+		tests = append(tests, signedURLTest{"putURLEnforcedAbsentContentType", putURLEnforcedAbsentContentType, "", true})
+		tests = append(tests, signedURLTest{"putURLEnforcedAbsentContentType", putURLEnforcedAbsentContentType, differentContentType, false})
 	}
 	if putURLWithoutContentType != "" {
-		tests = append(tests, signedURLTest{http.MethodPut, "", putURLWithoutContentType, true})
+		tests = append(tests, signedURLTest{"putURLWithoutContentType", putURLWithoutContentType, "", true})
+		// From the SignedURLOptions docstring:
+		// If EnforceAbsentContentType is false and ContentType is the empty string,
+		// then PUTing without a Content-Type header will succeed, but it is
+		// implementation-specific whether providing a Content-Type header will fail.
+		// --> so, we don't have a test for putURLWithoutContentType and a non-empty ContentType.
 	}
 	for _, test := range tests {
 		req, err := http.NewRequest(http.MethodPut, test.url, strings.NewReader(contents))
 		if err != nil {
-			t.Fatalf("failed to create PUT HTTP request using %s URL (content-type=%q): %v", test.urlMethod, test.contentType, err)
+			t.Fatalf("failed to create PUT HTTP request using %q: %v", test.urlDescription, err)
 		}
 		if test.contentType != "" {
 			req.Header.Set("Content-Type", test.contentType)
 		}
 		if resp, err := client.Do(req); err != nil {
-			t.Fatalf("PUT failed with %s URL (content-type=%q): %v", test.urlMethod, test.contentType, err)
+			t.Fatalf("PUT to %q with ContentType %q failed: %v", test.urlDescription, test.contentType, err)
 		} else {
 			defer resp.Body.Close()
 			success := resp.StatusCode >= 200 && resp.StatusCode < 300
 			if success != test.wantSuccess {
-				t.Errorf("PUT with %s URL (content-type=%q) got status code %d, want 2xx? %v", test.urlMethod, test.contentType, resp.StatusCode, test.wantSuccess)
+				t.Errorf("PUT to %q with ContentType %q got status code %d, wanted 2xx? %v", test.urlDescription, test.contentType, resp.StatusCode, test.wantSuccess)
 				gotBody, _ := ioutil.ReadAll(resp.Body)
 				t.Errorf(string(gotBody))
 			}
@@ -2167,30 +2275,30 @@ func testSignedURL(t *testing.T, newHarness HarnessMaker) {
 
 	// GET it. Try with all URLs, only getURL should work.
 	for _, test := range []struct {
-		urlMethod   string
-		url         string
-		wantSuccess bool
+		urlDescription string
+		url            string
+		wantSuccess    bool
 	}{
-		{http.MethodDelete, deleteURL, false},
-		{http.MethodPut, putURLWithoutContentType, false},
-		{http.MethodGet, getURLNoParams, false},
-		{http.MethodGet, getURL, true},
+		{"deleteURL", deleteURL, false},
+		{"putURLWithoutContentType", putURLWithoutContentType, false},
+		{"getURLNoParams", getURLNoParams, false},
+		{"getURL", getURL, true},
 	} {
 		if resp, err := client.Get(test.url); err != nil {
-			t.Fatalf("GET with %s URL failed: %v", test.urlMethod, err)
+			t.Fatalf("GET with %s URL failed: %v", test.urlDescription, err)
 		} else {
 			defer resp.Body.Close()
 			success := resp.StatusCode >= 200 && resp.StatusCode < 300
 			if success != test.wantSuccess {
-				t.Errorf("GET with %s URL got status code %d, want 2xx? %v", test.urlMethod, resp.StatusCode, test.wantSuccess)
+				t.Errorf("GET to %q got status code %d, want 2xx? %v", test.urlDescription, resp.StatusCode, test.wantSuccess)
 				gotBody, _ := ioutil.ReadAll(resp.Body)
 				t.Errorf(string(gotBody))
 			} else if success {
 				gotBody, err := ioutil.ReadAll(resp.Body)
 				if err != nil {
-					t.Errorf("GET with %s URL failed to read response body: %v", test.urlMethod, err)
+					t.Errorf("GET to %q failed to read response body: %v", test.urlDescription, err)
 				} else if gotBodyStr := string(gotBody); gotBodyStr != contents {
-					t.Errorf("GET with %s URL got body %q, want %q", test.urlMethod, gotBodyStr, contents)
+					t.Errorf("GET to %q got body %q, want %q", test.urlDescription, gotBodyStr, contents)
 				}
 			}
 		}
@@ -2198,25 +2306,25 @@ func testSignedURL(t *testing.T, newHarness HarnessMaker) {
 
 	// DELETE it. Try with all URLs, only deleteURL should work.
 	for _, test := range []struct {
-		urlMethod   string
-		url         string
-		wantSuccess bool
+		urlDescription string
+		url            string
+		wantSuccess    bool
 	}{
-		{http.MethodGet, getURL, false},
-		{http.MethodPut, putURLWithoutContentType, false},
-		{http.MethodDelete, deleteURL, true},
+		{"getURL", getURL, false},
+		{"putURLWithoutContentType", putURLWithoutContentType, false},
+		{"deleteURL", deleteURL, true},
 	} {
 		req, err := http.NewRequest(http.MethodDelete, test.url, nil)
 		if err != nil {
-			t.Fatalf("failed to create DELETE HTTP request using %s URL: %v", test.urlMethod, err)
+			t.Fatalf("failed to create DELETE HTTP request using %q: %v", test.urlDescription, err)
 		}
 		if resp, err := client.Do(req); err != nil {
-			t.Fatalf("DELETE with %s URL failed: %v", test.urlMethod, err)
+			t.Fatalf("DELETE to %q failed: %v", test.urlDescription, err)
 		} else {
 			defer resp.Body.Close()
 			success := resp.StatusCode >= 200 && resp.StatusCode < 300
 			if success != test.wantSuccess {
-				t.Fatalf("DELETE with %s URL got status code %d, want 2xx? %v", test.urlMethod, resp.StatusCode, test.wantSuccess)
+				t.Fatalf("DELETE to %q got status code %d, want 2xx? %v", test.urlDescription, resp.StatusCode, test.wantSuccess)
 				gotBody, _ := ioutil.ReadAll(resp.Body)
 				t.Errorf(string(gotBody))
 			}
@@ -2343,6 +2451,13 @@ func testAs(t *testing.T, newHarness HarnessMaker, st AsTest) {
 	} else {
 		defer func() { _ = b.Delete(ctx, copyKey) }()
 	}
+
+	for _, method := range []string{http.MethodGet, http.MethodPut, http.MethodDelete} {
+		_, err = b.SignedURL(ctx, key, &blob.SignedURLOptions{Method: method, BeforeSign: st.BeforeSign})
+		if err != nil && gcerrors.Code(err) != gcerrors.Unimplemented {
+			t.Errorf("got err %v when signing url with method %q", err, method)
+		}
+	}
 }
 
 func benchmarkRead(b *testing.B, bkt *blob.Bucket) {
